/*
 *  Version: 1.0
 *
 *  The contents of this file are subject to the OpenVPMS License Version
 *  1.0 (the 'License'); you may not use this file except in compliance with
 *  the License. You may obtain a copy of the License at
 *  http://www.openvpms.org/license/
 *
 *  Software distributed under the License is distributed on an 'AS IS' basis,
 *  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 *  for the specific language governing rights and limitations under the
 *  License.
 *
 *  Copyright 2006 (C) OpenVPMS Ltd. All Rights Reserved.
 *
 *  $Id: AbstractLayoutStrategy.java 4057 2011-02-03 03:30:17Z tanderson $
 */

package org.openvpms.web.component.im.layout;

import echopointng.TabbedPane;
import echopointng.tabbedpane.TabModel;
import nextapp.echo2.app.Column;
import nextapp.echo2.app.Component;
import nextapp.echo2.app.Grid;
import nextapp.echo2.app.Label;
import nextapp.echo2.app.SelectField;
import org.openvpms.component.business.domain.im.archetype.descriptor.ArchetypeDescriptor;
import org.openvpms.component.business.domain.im.archetype.descriptor.NodeDescriptor;
import org.openvpms.component.business.domain.im.common.IMObject;
import org.openvpms.component.business.service.archetype.helper.DescriptorHelper;
import org.openvpms.web.component.focus.FocusGroup;
import org.openvpms.web.component.focus.FocusHelper;
import org.openvpms.web.component.im.filter.ChainedNodeFilter;
import org.openvpms.web.component.im.filter.FilterHelper;
import org.openvpms.web.component.im.filter.NodeFilter;
import org.openvpms.web.component.im.view.ComponentState;
import org.openvpms.web.component.im.view.IMObjectComponentFactory;
import org.openvpms.web.component.property.Property;
import org.openvpms.web.component.property.PropertySet;
import org.openvpms.web.component.util.ColumnFactory;
import org.openvpms.web.component.util.GridFactory;
import org.openvpms.web.component.util.LabelFactory;
import org.openvpms.web.component.util.RowFactory;
import org.openvpms.web.component.util.TabPaneModel;
import org.openvpms.web.component.util.TabbedPaneFactory;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * Abstract implementation of the {@link IMObjectLayoutStrategy} interface.
 *
 * @author <a href="mailto:support@openvpms.org">OpenVPMS Team</a>
 * @version $LastChangedDate: 2011-02-03 13:30:17 +1000 (Thu, 03 Feb 2011) $
 */
public abstract class AbstractLayoutStrategy implements IMObjectLayoutStrategy {

    /**
     * List of component states, used to determine initial focus.
     */
    private final List<ComponentState> components
            = new ArrayList<ComponentState>();

    /**
     * Pre-created component states, keyed on property name.
     */
    private Map<String, ComponentState> states = new HashMap<String, ComponentState>();

    /**
     * The focus group of the current component.
     */
    private FocusGroup focusGroup;

    /**
     * If <tt>true</tt> keep layout state after invoking <tt>apply()</tt>. Use this if the same strategy will be used
     * to layout a component multiple times.
     */
    private boolean keepState;

    /**
     * Sanity checker to detect recursion.
     */
    boolean inApply;


    /**
     * Constructs a <tt>AbstractLayoutStrategy</tt>.
     */
    public AbstractLayoutStrategy() {
        this(false);
    }

    /**
     * Constructs a <tt>AbstractLayoutStrategy</tt>.
     *
     * @param keepState if <tt>true</tt> keep layout state. Use this if the same strategy will be used to layout a
     *                  component multiple times
     */
    public AbstractLayoutStrategy(boolean keepState) {
        this.keepState = keepState;
    }

    /**
     * Pre-registers a component for inclusion in the layout.
     * <p/>
     * The component must be associated with a property.
     *
     * @param state the component state
     * @throws IllegalStateException if the component isn't associated with a property
     */
    public void addComponent(ComponentState state) {
        Property property = state.getProperty();
        if (property == null) {
            throw new IllegalArgumentException("Argument 'state' must be associated with a property");
        }
        states.put(property.getName(), state);
    }

    /**
     * Apply the layout strategy.
     * <p/>
     * This renders an object in a <code>Component</code>, using a factory to
     * create the child components.
     *
     * @param object     the object to apply
     * @param properties the object's properties
     * @param parent     the parent object. May be <tt>null</tt>
     * @param context    the layout context
     * @return the component containing the rendered <code>object</code>
     */
    public ComponentState apply(IMObject object, PropertySet properties,
                                IMObject parent, LayoutContext context) {
        ComponentState state;
        if (inApply) {
            throw new IllegalStateException("Cannot call apply() recursively");
        }
        inApply = true;
        try {
            focusGroup = new FocusGroup(DescriptorHelper.getDisplayName(object));
            Column column = ColumnFactory.create("CellSpacing");
            doLayout(object, properties, parent, column, context);
            focusGroup.setDefault(getDefaultFocus());
            state = new ComponentState(column, focusGroup);
            components.clear();
            if (!keepState) {
                focusGroup = null;
            }
        } finally {
            inApply = false;
        }
        return state;
    }

    /**
     * Returns the focus group.
     *
     * @return the focus group, or <code>null</code> if it hasn't been
     *         initialised
     */
    protected FocusGroup getFocusGroup() {
        return focusGroup;
    }

    /**
     * Lay out out the object in the specified container.
     *
     * @param object     the object to lay out
     * @param properties the object's properties
     * @param parent     the parent object. May be <tt>null</tt>
     * @param container  the container to use
     * @param context    the layout context
     */
    protected void doLayout(IMObject object, PropertySet properties, IMObject parent,
                            Component container, LayoutContext context) {
        ArchetypeDescriptor archetype = context.getArchetypeDescriptor(object);
        List<NodeDescriptor> simple = getSimpleNodes(archetype);
        List<NodeDescriptor> complex = getComplexNodes(archetype);

        NodeFilter filter = getNodeFilter(object, context);
        simple = filter(object, simple, filter);
        complex = filter(object, complex, filter);

        doSimpleLayout(object, parent, simple, properties, container, context);
        doComplexLayout(object, parent, complex, properties, container, context);
    }

    /**
     * Lays out child components in a grid.
     *
     * @param object      the object to lay out
     * @param parent      the parent object. May be <tt>null</tt>
     * @param descriptors the property descriptors
     * @param properties  the properties
     * @param container   the container to use
     * @param context     the layout context
     */
    protected void doSimpleLayout(IMObject object, IMObject parent, List<NodeDescriptor> descriptors,
                                  PropertySet properties, Component container,
                                  LayoutContext context) {
        if (!descriptors.isEmpty()) {
            Grid grid = createGrid(descriptors);
            doGridLayout(object, descriptors, properties, grid, context);
            container.add(ColumnFactory.create("Inset.Small", grid));
        }
    }

    /**
     * Lays out each child component in a tabbed pane.
     *
     * @param object      the object to lay out
     * @param parent      the parent object. May be <tt>null</tt>
     * @param descriptors the property descriptors
     * @param properties  the properties
     * @param container   the container to use
     * @param context     the layout context
     */
    protected void doComplexLayout(IMObject object, IMObject parent, List<NodeDescriptor> descriptors,
                                   PropertySet properties, Component container,
                                   LayoutContext context) {
        if (!descriptors.isEmpty()) {
            TabModel model = doTabLayout(object, descriptors, properties, container, context, false);
            TabbedPane pane = TabbedPaneFactory.create(model);

            pane.setSelectedIndex(0);
            container.add(pane);
        }
    }

    /**
     * Returns the 'simple' nodes.
     *
     * @param archetype the archetype
     * @return the simple nodes
     * @see org.openvpms.component.business.domain.im.archetype.descriptor.ArchetypeDescriptor#getSimpleNodeDescriptors()
     */
    protected List<NodeDescriptor> getSimpleNodes(
            ArchetypeDescriptor archetype) {
        return archetype.getSimpleNodeDescriptors();
    }

    /**
     * Returns the 'complex' nodes.
     *
     * @param archetype the archetype
     * @return the complex nodes
     * @see org.openvpms.component.business.domain.im.archetype.descriptor.ArchetypeDescriptor#getComplexNodeDescriptors()
     */
    protected List<NodeDescriptor> getComplexNodes(
            ArchetypeDescriptor archetype) {
        return archetype.getComplexNodeDescriptors();
    }

    /**
     * Returns a node filter to filter nodes. This implementation return {@link
     * LayoutContext#getDefaultNodeFilter()}.
     *
     * @param object  the object to filter nodes for
     * @param context the context
     * @return a node filter to filter nodes, or <code>null</code> if no
     *         filterering is required
     */
    protected NodeFilter getNodeFilter(IMObject object, LayoutContext context) {
        return context.getDefaultNodeFilter();
    }

    /**
     * Helper to create a chained node filter from the default node filter and a
     * custom node filter.
     *
     * @param context the context
     * @param filter  the node filter
     * @return a new chained node filter
     */
    protected ChainedNodeFilter getNodeFilter(LayoutContext context,
                                              NodeFilter filter) {
        return FilterHelper.chain(context.getDefaultNodeFilter(), filter);
    }

    /**
     * Filters a set of node descriptors, using the specfied node filter.
     *
     * @param object      the object
     * @param descriptors the node descriptors to filter
     * @param filter      the filter to use
     * @return the filtered nodes
     */
    protected List<NodeDescriptor> filter(IMObject object,
                                          List<NodeDescriptor> descriptors,
                                          NodeFilter filter) {
        return FilterHelper.filter(object, filter, descriptors);
    }

    /**
     * Lays out child components in columns.
     *
     * @param object      the parent object
     * @param descriptors the property descriptors
     * @param properties  the properties
     * @param grid        the grid to use
     * @param context     the layout context
     */
    protected void doGridLayout(IMObject object,
                                List<NodeDescriptor> descriptors,
                                PropertySet properties, Grid grid,
                                LayoutContext context) {
        ComponentSet set = createComponentSet(object, descriptors, properties,
                                              context);
        ComponentState[] states = set.getComponents().toArray(new ComponentState[set.getComponents().size()]);
        Component[] components = new Component[states.length];
        String[] labels = new String[states.length];
        for (int i = 0; i < states.length; ++i) {
            ComponentState state = states[i];
            Component component = state.getComponent();
            components[i] = component;
            labels[i] = set.getLabel(state);
            setFocusTraversal(state);
            if (component instanceof SelectField) {
                // workaround for render bug in firefox. See OVPMS-239
                components[i] = RowFactory.create(component);
            }
        }
        int size = components.length;
        int columns = (grid.getSize() <= 2) ? 1 : 2;
        int rows;
        if (columns == 1) {
            rows = size;
        } else {
            rows = (size / 2) + (size % 2);
        }
        for (int i = 0, j = rows; i < rows; ++i, ++j) {
            add(grid, labels[i], components[i]);
            if (j < size) {
                add(grid, labels[j], components[j]);
            }
        }
    }

    /**
     * Lays out a component grid.
     *
     * @param grid      the grid
     * @param container the container to add the grid to
     */
    protected void doGridLayout(ComponentGrid grid, Component container) {
        Grid g = GridFactory.create(grid.getColumns() * 2);
        for (int row = 0; row < grid.getRows(); ++row) {
            for (int col = 0; col < grid.getColumns(); ++col) {
                ComponentState state = grid.get(row, col);
                if (state != null) {
                    setFocusTraversal(state);
                    Component component = state.getComponent();
                    if (component instanceof SelectField) {
                        // workaround for render bug in firefox. See OVPMS-239
                        component = RowFactory.create(component);
                    }
                    add(g, state.getDisplayName(), component);
                } else {
                    add(g, null, LabelFactory.create());
                }
            }
        }
        container.add(g);
    }

    /**
     * Lays out child components in a tab model.
     *
     * @param object       the parent object
     * @param descriptors  the property descriptors
     * @param properties   the properties
     * @param container    the container
     * @param context      the layout context
     * @param shortcutHint a hint to display short cuts for tabs. If <tt>false</tt> shortcuts will only be displayed if
     *                     there is more than one descriptor. Shortcuts will never be displayed if the layout depth
     *                     is non-zero
     * @return the tab model
     */
    protected TabPaneModel doTabLayout(IMObject object, List<NodeDescriptor> descriptors, PropertySet properties,
                                       Component container, LayoutContext context, boolean shortcutHint) {
        TabPaneModel model;
        boolean shortcuts = false;
        if (context.getLayoutDepth() == 0 && (descriptors.size() > 1 || shortcutHint)) {
            model = createTabModel(container);
            shortcuts = true;
        } else {
            model = createTabModel(null);
        }
        doTabLayout(object, descriptors, properties, model, context, shortcuts);
        return model;
    }

    /**
     * Creates a new tab model.
     *
     * @param container the tab container. May be <tt>null</tt>
     * @return a new tab model
     */
    protected TabPaneModel createTabModel(Component container) {
        return new TabPaneModel(container);
    }

    /**
     * Lays out child components in a tab model.
     *
     * @param object      the parent object
     * @param descriptors the property descriptors
     * @param properties  the properties
     * @param model       the tab model
     * @param context     the layout context
     * @param shortcuts   if <tt>true</tt> include short cuts
     */
    protected void doTabLayout(IMObject object, List<NodeDescriptor> descriptors, PropertySet properties,
                               TabPaneModel model, LayoutContext context, boolean shortcuts) {
        int shortcut = 1;
        for (NodeDescriptor nodeDesc : descriptors) {
            Property property = properties.get(nodeDesc);
            ComponentState child = createComponent(property, object, context);
            Component inset = ColumnFactory.create("Inset", child.getComponent());
            setFocusTraversal(child);
            String text = child.getDisplayName();
            if (text == null) {
                text = nodeDesc.getDisplayName();
            }

            if (shortcuts && shortcut <= 10) {
                text = getShortcut(text, shortcut);
                ++shortcut;
            }
            model.addTab(text, inset);
        }
    }

    /**
     * Adds a tab to a tab model.
     *
     * @param model       the tab  model
     * @param property    property
     * @param component   the component to add
     * @param addShortcut if <tt>true</tt> add a tab shortcut
     */
    protected void addTab(TabPaneModel model, Property property, ComponentState component, boolean addShortcut) {
        setFocusTraversal(component);
        String text = component.getDisplayName();
        if (text == null) {
            text = property.getDisplayName();
        }
        if (addShortcut && model.size() < 10) {
            text = getShortcut(text, model.size() + 1);
        }
        Component inset = ColumnFactory.create("Inset", component.getComponent());
        model.addTab(text, inset);
    }

    /**
     * Creates a set of components to be rendered from the supplied descriptors.
     *
     * @param object      the parent object
     * @param descriptors the property descriptors
     * @param properties  the properties
     * @param context     the layout context
     * @return the components
     */
    protected ComponentSet createComponentSet(IMObject object,
                                              List<NodeDescriptor> descriptors,
                                              PropertySet properties,
                                              LayoutContext context) {
        ComponentSet result = new ComponentSet();
        for (NodeDescriptor descriptor : descriptors) {
            Property property = properties.get(descriptor);
            ComponentState component = createComponent(property, object,
                                                       context);
            String displayName = component.getDisplayName();
            if (displayName == null) {
                displayName = descriptor.getDisplayName();
            }
            result.add(component, displayName);
        }
        return result;
    }

    /**
     * Helper to add a node to a container.
     *
     * @param container the container
     * @param name      the node display name. May be <tt>null</tt>
     * @param component the component representing the node
     */
    protected void add(Component container, String name, Component component) {
        Label label = LabelFactory.create();
        if (name != null) {
            label.setText(name);
        }
        container.add(label);
        container.add(component);
    }

    /**
     * Helper to add a node to a container, setting its tab index.
     *
     * @param container the container
     * @param name      the node display name
     * @param component the component representing the node
     */
    protected void add(Component container, String name,
                       ComponentState component) {
        add(container, name, component.getComponent());
        setFocusTraversal(component);
    }

    /**
     * Helper to add a property to a container, setting its tab index.
     *
     * @param container the container
     * @param component the component representing the property
     */
    protected void add(Component container, ComponentState component) {
        Property property = component.getProperty();
        String name = null;
        if (property != null) {
            name = property.getDisplayName();
        }
        add(container, name, component);
        setFocusTraversal(component);
    }

    /**
     * Creates a component for a property.
     * <p/>
     * If there is a pre-existing component, registered via {@link #addComponent}, this will be returned.
     *
     * @param property the property
     * @param parent   the parent object
     * @param context  the layout context
     * @return a component to display <code>property</code>
     */
    protected ComponentState createComponent(Property property, IMObject parent, LayoutContext context) {
        ComponentState result = states.get(property.getName());
        if (result == null) {
            IMObjectComponentFactory factory = context.getComponentFactory();
            result = factory.create(property, parent);
        }
        return result;
    }

    /**
     * Creates a grid with the no. of columns determined by the no. of
     * node descriptors.
     *
     * @param descriptors the node descriptors
     * @return a new grid
     */
    protected Grid createGrid(List<NodeDescriptor> descriptors) {
        return (descriptors.size() <= 4) ? GridFactory.create(2)
                                         : GridFactory.create(4);
    }

    /**
     * Returns the default focus component.
     * <p/>
     * Delegates to {@link #getDefaultFocus(List <ComponentState>)}.
     *
     * @return the default focus component, or <tt>null</tt> if none is found
     */
    protected Component getDefaultFocus() {
        return getDefaultFocus(components);
    }

    /**
     * Returns the default focus component.
     * <p/>
     * This implementation returns the first focusable component.
     *
     * @param components the components
     * @return the default focus component, or <tt>null</tt> if none is found
     */
    protected Component getDefaultFocus(List<ComponentState> components) {
        return getFocusable(components);
    }

    /**
     * Returns the first focusable component, selecting invalid properties
     * in preference to other components.
     *
     * @param components the components
     * @return the first focusable component
     */
    protected Component getFocusable(List<ComponentState> components) {
        Component result = null;
        for (ComponentState state : components) {
            Component child = state.getFocusable();
            if (child != null) {
                Property property = state.getProperty();
                if (property != null && !property.isValid()) {
                    result = child;
                    break;
                }
                if (result == null) {
                    result = child;
                }
            }
        }
        return result;
    }

    /**
     * Sets the focus traversal index of a component, if it is a focus traversal
     * participant.
     * NOTE: if a component doesn't specify a focus group,
     * this may register a child component with the focus group rather than the
     * parent.
     *
     * @param state the component state
     */
    protected void setFocusTraversal(ComponentState state) {
        Component component = state.getComponent();
        if (state.getFocusGroup() != null) {
            focusGroup.add(state.getFocusGroup());
            components.add(state);
        } else {
            Component focusable = FocusHelper.getFocusable(component);
            if (focusable != null) {
                focusGroup.add(focusable);
                components.add(state);
            }
        }
    }

    /**
     * Returns a shortcut for a tab.
     * Shortcuts no.s must be from 1..10, and will be displayed as '1..9, 0'.
     *
     * @param name     the tab name
     * @param shortcut the shortcut no.
     * @return the shortcut text
     */
    protected String getShortcut(String name, int shortcut) {
        if (shortcut == 10) {
            shortcut = 0;
        }
        return "&" + shortcut + " " + name;
    }

}

